/*  $Header: /home/cvsroot/dvipdfmx/src/dpxfile.c,v 1.8 2004/09/12 14:35:41 hirata Exp $
    
    This is dvipdfmx, an eXtended version of dvipdfm by Mark A. Wicks.

    Copyright (C) 2002 by Jin-Hwan Cho and Shunsaku Hirata,
    the dvipdfmx project team <dvipdfmx@project.ktug.or.kr>
    
    Copyright (C) 1998, 1999 by Mark A. Wicks <mwicks@kettering.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include "system.h"
#include "mem.h"
#include "error.h"
#include "dpxutil.h"
#include "mfileio.h"

#include "dpxfile.h"

static FILE *dpx_open_fontmap_file  (const char *filename);
static FILE *dpx_open_sfd_file      (const char *filename);
static FILE *dpx_open_cmap_file     (const char *filename);
static FILE *dpx_open_enc_file      (const char *filename);
static FILE *dpx_open_opentype_file (const char *filename);
static FILE *dpx_open_truetype_file (const char *filename);
static FILE *dpx_open_type1_file    (const char *filename);
static FILE *dpx_open_iccp_file     (const char *filename);

FILE *
dpx_open_file (const char *filename, int type)
{
  FILE *fp = NULL;

  switch (type) {
  case DPX_RES_TYPE_FONTMAP:
    fp = dpx_open_fontmap_file(filename);
    break;
  case DPX_RES_TYPE_T1FONT:
    fp = dpx_open_type1_file(filename);
    break;
  case DPX_RES_TYPE_TTFONT:
    fp = dpx_open_truetype_file(filename);
    break;
  case DPX_RES_TYPE_OTFONT:
    fp = dpx_open_opentype_file(filename);
    break;
  case DPX_RES_TYPE_PKFONT:
    break;
  case DPX_RES_TYPE_CMAP:
    fp = dpx_open_cmap_file(filename);
    break;
  case DPX_RES_TYPE_ENC:
    fp = dpx_open_enc_file(filename);
    break;
  case DPX_RES_TYPE_SFD:
    fp = dpx_open_sfd_file(filename);
    break;
  case DPX_RES_TYPE_ICCPROFILE:
    fp = dpx_open_iccp_file(filename);
    break;
  default:
    ERROR("Unknown resource type: %d", type);
    break;
  }

  return fp;
}

static FILE *
dpx_open_iccp_file (const char *filename)
{
  FILE  *fp = NULL;
  static const char *suffixes[] = {
    ".icc", ".icm", NULL
  };
  char  *fullname, *altname, *suffix;
  int    i;

  fullname = kpse_find_file(filename, kpse_program_binary_format, 0);
  if (fullname) {
    fp = MFOPEN(fullname, FOPEN_RBIN_MODE);
    RELEASE(fullname);
    return fp;
  }

  suffix = strrchr(filename, '.');
  if (!suffix) {
    for (i = 0; fp == NULL &&
	   suffixes[i] != NULL; i++) {
      altname = NEW(strlen(filename)+strlen(suffixes[i])+1, char);
      strcpy(altname, filename);
      strcat(altname, suffixes[i]);
      fullname = kpse_find_file(altname, kpse_program_binary_format, 0);
      if (fullname) {
	fp = MFOPEN(fullname, FOPEN_RBIN_MODE);
	RELEASE(fullname);
      }
      RELEASE(altname);
    }
  }

  return fp;
}

#ifndef HAVE_KPSE_ENC_FORMATS
#define kpse_enc_format  kpse_tex_ps_header_format
#define kpse_cmap_format kpse_program_text_format
#define kpse_sfd_format  kpse_program_text_format
#endif

static FILE *
dpx_open_fontmap_file (const char *filename)
{
  FILE *fp;
  char *fullname, *altname, *suffix;

  suffix = strrchr(filename, '.');
  if (!suffix || strcmp(suffix, ".map") != 0) {
    altname = NEW(strlen(filename)+strlen(".map")+1, char);    
    strcpy(altname, filename);
    strcat(altname, ".map");
  } else {
    altname = NULL;
  }

#ifdef MIKTEX
  if (!miktex_find_app_input_file("dvipdfm", filename, work_buffer)) {
    if (altname &&
	miktex_find_app_input_file("dvipdfm", altname, work_buffer)) {
      fullname = work_buffer;
    } else {
      fullname = NULL;
    }
  } else {
    fullname = work_buffer;
  }
#else /* !MIKTEX */

  fullname = kpse_find_file(filename, kpse_fontmap_format, 0);

#ifndef HAVE_KPSE_ENC_FORMATS
  if (!fullname && altname)
    fullname = kpse_find_file(altname, kpse_fontmap_format, 0);
  if (!fullname)
    fullname = kpse_find_file(filename, kpse_program_text_format, 0);
  if (!fullname && altname)
    fullname = kpse_find_file(altname,  kpse_program_text_format, 0);
#endif /* !HAVE_KPSE_ENC_FORMATS */

#endif /* MIKETEX */

  if (altname)
    RELEASE(altname);

  if (!fullname)
    return NULL;

  fp = MFOPEN(fullname, FOPEN_R_MODE);

#ifndef MIKTEX
  RELEASE(fullname);
#endif

  return fp;
}

static FILE *
dpx_open_cmap_file (const char *filename)
{
  FILE *fp;
  char *fullname;

#ifdef MIKTEX
  if (!miktex_find_app_input_file("dvipdfm", filename, work_buffer)) {
    if (miktex_get_acrobat_font_dir(work_buffer)) {
      strcat(work_buffer, "\\..\\CMap\\");
      strcat(work_buffer, filename);
      fullname = work_buffer;
    } else {
      fullname = NULL;
    }
  } else {
    fullname = work_buffer;
  }
#else

  fullname = kpse_find_file(filename, kpse_cmap_format, 0);

#ifndef HAVE_KPSE_ENC_FORMATS
  if (!fullname) {
    char *altname;
    char *suffix;

    suffix = strrchr(filename, '.');
    if (suffix == NULL || strcmp(suffix, ".cmap") != 0) {
      altname = NEW(strlen(filename)+strlen(".cmap")+1, char);
      strcpy(altname, filename);
      strcat(altname, ".cmap");
    }  else {
      altname = NULL;
    }
    if (altname)
      fullname = kpse_find_file(altname, kpse_cmap_format, 0);
    if (!fullname) {
      kpse_reset_program_name("cmap");
      fullname = kpse_find_file(filename, kpse_cmap_format, 0);
      if (!fullname && altname)
	fullname = kpse_find_file(altname, kpse_cmap_format, 0);
      kpse_reset_program_name("dvipdfm");
    }
    if (altname)
      RELEASE(altname);
  }
#endif /* !KPSE_ENC_FORMATS */

#endif /* MIKTEX */

  if (fullname == NULL)
    return NULL;

  fp = MFOPEN(fullname, FOPEN_RBIN_MODE);

#ifndef MIKTEX
  RELEASE(fullname); /* kpse malloc'ed */
#endif

  return fp;
}


static FILE *
dpx_open_sfd_file (const char *sfdname)
{
  FILE *fp;
  char *filename, *fullname;

  filename = NEW(strlen(sfdname)+strlen(".sfd")+1, char);
  sprintf(filename, "%s.sfd", sfdname);

#ifdef MIKTEX
  if (!miktex_find_app_input_file("ttf2pk", filename, work_buffer))
    fullname = NULL;
  else {
    fullname = work_buffer;
  }
#else /* !MIKTEX */

#ifndef HAVE_KPSE_ENC_FORMATS
/* Change the program name temporarily to find SFD files from
 * the texmf/ttf2pk directory (contributed by Akira Kakuto).
 */

  kpse_reset_program_name("ttf2pk");
#endif /* HAVE_KPSE_ENC_FORMATS */

  fullname = kpse_find_file(filename, kpse_sfd_format, 0);

#ifndef HAVE_KPSE_ENC_FORMATS
  kpse_reset_program_name("dvipdfm");
#endif /* HAVE_KPSE_ENC_FORMATS */

#endif /* MIKTEX */

  RELEASE(filename);

  if (!fullname)
    return NULL;

  fp = MFOPEN(fullname, FOPEN_R_MODE);

#ifndef MIKTEX
  RELEASE(fullname);
#endif
  
  return fp;
}

static FILE *
dpx_open_enc_file (const char *filename)
{
  FILE *fp;
  char *fullname, *altname, *suffix;

  suffix = strrchr(filename, '.');
  if (!suffix || strcmp(suffix, ".enc") != 0) {
    altname = NEW(strlen(filename)+strlen(".enc")+1, char);    
    sprintf(altname, "%s.enc", filename);
  } else {
    altname = NULL;
  }

#ifdef MIKTEX
  if (!miktex_find_psheader_file(filename, work_buffer) &&
      !miktex_find_app_input_file("dvipdfm", filename, work_buffer)) {
    if (altname) {
      if (!miktex_find_psheader_file(altname, work_buffer) &&
	  !miktex_find_app_input_file("dvipdfm", altname, work_buffer))
	fullname = NULL;
      else {
	fullname = work_buffer;
      }
    } else {
      fullname = NULL;
    }
  } else {
    fullname = work_buffer;
  }
#else /* !MIKETEX */
  fullname = kpse_find_file(filename, kpse_enc_format, 0);
#ifndef HAVE_KPSE_ENC_FORMATS
  if (!fullname) {
    if (altname) 
      fullname = kpse_find_file(altname,  kpse_enc_format, 0);
    if (!fullname) 
      fullname = kpse_find_file(filename, kpse_program_text_format, 0);
    if (!fullname && altname)
      fullname = kpse_find_file(altname,  kpse_program_text_format, 0);
  }
#endif /* !HAVE_KPSE_ENC_FORMATS */
#endif /* MIKETEX */

  if (altname)
    RELEASE(altname);

  if (!fullname)
    return NULL;

  fp = MFOPEN(fullname, FOPEN_R_MODE);

#ifndef MIKTEX
  RELEASE(fullname); /* kpse malloc'ed */
#endif

  return fp;
}

static FILE *
dpx_open_type1_file (const char *filename)
{
  FILE  *fp;
  char  *fullname;


  fullname = kpse_find_file(filename, kpse_type1_format, 0);
  if (!fullname)
    return NULL;

  fp = MFOPEN(fullname, FOPEN_RBIN_MODE);
  RELEASE(fullname);

  return fp;
}

static FILE *
dpx_open_truetype_file (const char *filename)
{
  FILE  *fp;
  char  *fullname;


  fullname = kpse_find_file(filename, kpse_truetype_format, 0);
  if (!fullname)
    return NULL;

  fp = MFOPEN(fullname, FOPEN_RBIN_MODE);
  RELEASE(fullname);

  return fp;
}

#ifndef HAVE_KPSE_OPENTYPE_FORMAT
#define kpse_opentype_format kpse_program_binary_format
#endif
static FILE *
dpx_open_opentype_file (const char *filename)
{
  FILE *fp;
  char *altname, *fullname, *suffix;

  if (!filename)
    return NULL;

  altname = NEW(strlen(filename)+5, char);
  memset(altname, 0, strlen(filename)+strlen(".otf")+1);
  strcpy(altname, filename);
  suffix  = strrchr(filename, '.');
  if (!suffix || strcmp(suffix, ".otf"))
    strcat(altname, ".otf");

#ifdef MIKTEX
  fullname = work_buffer;
  if (!miktex_find_app_input_file("dvipdfm", altname, fullname)) {
    RELEASE(altname);
    return NULL;
  }
#else
  fullname = kpse_find_file(altname, kpse_opentype_format, 0);
  if (!fullname) {
    RELEASE(altname);
    return NULL;
  }
#endif
  RELEASE(altname);

  fp = MFOPEN(fullname, FOPEN_RBIN_MODE);

#ifndef MIKTEX
  RELEASE(fullname); /* kpse malloc'ed. */
#endif

  return fp;
}

