/*  $Header: /home/cvsroot/dvipdfmx/src/pdfdoc.c,v 1.35 2004/11/11 10:50:21 hirata Exp $
 
    This is dvipdfmx, an eXtended version of dvipdfm by Mark A. Wicks.

    Copyright (C) 2002 by Jin-Hwan Cho and Shunsaku Hirata,
    the dvipdfmx project team <dvipdfmx@project.ktug.or.kr>
    
    Copyright (C) 1998, 1999 by Mark A. Wicks <mwicks@kettering.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

/*
 * TODO: Many things...
 *  {begin,end}_{bead,article}, box stack, name tree (not limited to dests)...
 */
#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <time.h>

#include "system.h"
#include "mem.h"
#include "error.h"
#include "mfileio.h"

#include "numbers.h"

#include "pdfobj.h"
#include "pdfparse.h"
#include "pdfnames.h"

#include "pdfencrypt.h"

#include "pdfdev.h"
#include "pdfdraw.h"
#include "pdfcolor.h"

#include "pdfresource.h"
#include "pdffont.h"
#include "pdfximage.h"

#include "pdflimits.h"

#ifndef NO_THUMBNAIL
#include "pngimage.h"
#include "jpegimage.h"
#endif /* !NO_THUMBNAIL */

#include "pdfdoc.h"

#define PDFDOC_PAGES_ALLOC_SIZE   128u
#define PDFDOC_ARTICLE_ALLOC_SIZE 16
#define PDFDOC_BEAD_ALLOC_SIZE    16

static int verbose = 0;

#ifndef NO_THUMBNAIL
static char  thumb_enabled  = 0;
static char *thumb_basename = NULL;
static char  thumb_remove   = 0;

void
pdf_doc_enable_thumbnails (int t_remove)
{
#if HAVE_LIBPNG
  thumb_enabled = 1;
  thumb_remove  = t_remove;
#else
  WARN("Thumbnail support not available without PNG support.");
#endif
}

#define DEFAULT_TMPDIR "/tmp"
static char *
guess_thumb_name (const char *thumb_filename)
{
  /* Build path name for anticipated thumbnail image */
  char *tmpdir, *tmpname;

  if (!(tmpdir = getenv("TMP")) &&
      !(tmpdir = getenv("TEMP"))) 
    tmpdir = (char *) DEFAULT_TMPDIR;
  tmpname = NEW(strlen(tmpdir) +
		strlen(thumb_filename) + strlen(DIR_SEP_STRING) + 1,
		char);
  strcpy(tmpname, tmpdir);
  if (!IS_DIR_SEP(tmpname[strlen(tmpname)-1])) {
    strcat(tmpname, DIR_SEP_STRING);
  }
  strcat(tmpname, thumb_filename);

  return tmpname;
}

#ifndef HAVE_LIBPNG
static pdf_obj *
read_thumbnail (const char *thumb_filename) 
{
  WARN("Thumbnail support not available without PNG support.");
}
#else  /* HAVE_LIBPNG */
static pdf_obj *
read_thumbnail (const char *thumb_filename) 
{
  pdf_obj *image_ref;
  int      xobj_id;
  char    *guess_filename = NULL;
  int      found_in_cwd = 0;
  FILE    *fp;

  guess_filename = guess_thumb_name(thumb_filename);
  if ((fp = MFOPEN(thumb_filename, FOPEN_RBIN_MODE))) {
    found_in_cwd = 1;
  } else {
    fp = MFOPEN(guess_filename, FOPEN_RBIN_MODE);
  }

  if (!fp) {
    WARN("Could not open thumbnail file \"%s\"", thumb_filename);
    if (guess_filename)
      RELEASE(guess_filename);
    return NULL;
  } else if (!check_for_png(fp) && !check_for_jpeg(fp)) {
    WARN("Thumbnail \"%s\" not a png/jpeg file!", thumb_filename);
    MFCLOSE(fp);
    if (guess_filename)
      RELEASE(guess_filename);
    return NULL;
  }
  MFCLOSE(fp);

  if (found_in_cwd) {
    xobj_id = pdf_ximage_findresource(thumb_filename);
  } else {
    xobj_id = pdf_ximage_findresource(guess_filename);
  }

  if (xobj_id < 0) {
    WARN("Could not read thumbnail file \"%s\".", thumb_filename);
    image_ref = NULL;
  } else {
    image_ref = pdf_ximage_get_reference(xobj_id);
  }

  if (thumb_remove && found_in_cwd) {
    if (verbose > 1)
      MESG("\ndeleting thumbnail file \"%s\"...\n", thumb_filename);
    remove(thumb_filename);
  } else if (thumb_remove) {
    if (verbose > 1)
      MESG("\ndeleting thumbnail file \"%s\"...\n", guess_filename);
    remove(guess_filename);
  }

  if (guess_filename)
    RELEASE(guess_filename);

  return image_ref;
}

#endif /* HAVE_LIBPNG */
#endif /* !NO_THUMBNAIL */

void
pdf_doc_set_verbose (void)
{
  verbose++;
  pdf_font_set_verbose();
  pdf_color_set_verbose();
  pdf_ximage_set_verbose();
}

pdf_obj *docinfo = NULL;

static struct
{
  pdf_obj *catalog; /* Catalog dicrionary */

  /* Entries in Catalog dictionary */
  pdf_obj *viewerpref;
  pdf_obj *pagelabels;
  pdf_obj *pages;   /* Page tree root node */
  pdf_obj *names;   /* Name dictionary */
  pdf_obj *threads; /* articles_array */
} docroot = {NULL, NULL, NULL, NULL, NULL, NULL};

#define USE_MY_MEDIABOX (1 << 0)
struct page_entry
{
  pdf_obj  *page_obj;
  pdf_obj  *page_ref;

  int       flags;

  double    ref_x, ref_y;
  pdf_rect  cropbox;

  pdf_obj  *resources;

  /* Contents */
  pdf_obj  *background;
  pdf_obj  *contents;

  /* global bop, background, contents, global eop */
  pdf_obj  *content_refs[4];

  pdf_obj  *annots;
  pdf_obj  *beads;
};

static struct 
{
  pdf_obj *bop, *eop;
  pdf_rect mediabox;

  long num_pages; /* This is not actually total number of pages. */
  long max_pages;
  struct page_entry *pages;
} docpages = {
  NULL, NULL,
  {0.0, 0.0, 0.0, 0.0},

  0, 0,
  NULL
};

/*
 * Pages are starting at 1.
 * The page count does not increase until the page is finished.
 */
#define LASTPAGE()  (&(docpages.pages[docpages.num_pages]))
#define FIRSTPAGE() (&(docpages.pages[0]))
#define PAGECOUNT() (docpages.num_pages)
#define MAXPAGES()  (docpages.max_pages)

static void
resize_pages (long size)
{
  if (size > MAXPAGES()) {
    long i;

    docpages.pages = RENEW(docpages.pages, size, struct page_entry);
    for (i = docpages.max_pages; i < size; i++) {
      docpages.pages[i].page_obj   = NULL;
      docpages.pages[i].page_ref   = NULL;
      docpages.pages[i].flags      = 0;
      docpages.pages[i].resources  = NULL;
      docpages.pages[i].background = NULL;
      docpages.pages[i].contents   = NULL;
      docpages.pages[i].content_refs[0] = NULL; /* global bop */
      docpages.pages[i].content_refs[1] = NULL; /* background */
      docpages.pages[i].content_refs[2] = NULL; /* page body  */
      docpages.pages[i].content_refs[3] = NULL; /* global eop */
      docpages.pages[i].annots    = NULL;
      docpages.pages[i].beads     = NULL;
    }
    docpages.max_pages = size;
  }
}

static struct page_entry *
doc_page_entry (unsigned long page_no)
{
  struct page_entry *page;

  if (page_no > 65535ul)
    ERROR("Page number %ul too large!", page_no);
  else if (page_no == 0)
    ERROR("Invalid Page number %ul.", page_no);

  if (page_no > MAXPAGES()) {
    resize_pages(page_no + PDFDOC_PAGES_ALLOC_SIZE);
  }

  page = &(docpages.pages[page_no-1]);

  return page;
}


static void doc_close_page_tree  (void);

static void doc_init_names       (void);
static void doc_close_names      (void);
static void doc_init_articles    (void);
static void doc_close_articles   (void);
static void doc_init_bookmarks   (int bookmark_open_depth);
static void doc_close_bookmarks  (void);

void
pdf_doc_set_bop_content (const char *str, unsigned length)
{
  if (docpages.bop) {
    pdf_release_obj(docpages.bop);
  }
  if (length > 0) {
    docpages.bop = pdf_new_stream(STREAM_COMPRESS);
    pdf_add_stream(docpages.bop, str, length);
  } else {
    docpages.bop = NULL;
  }
}

void
pdf_doc_set_eop_content (const char *str, unsigned length)
{
  if (docpages.eop) {
    pdf_release_obj(docpages.eop);
  }
  if (length > 0) {
    docpages.eop = pdf_new_stream(STREAM_COMPRESS);
    pdf_add_stream(docpages.eop, str, length);
  } else {
    docpages.eop = NULL;
  }
}

/*
 * Docinfo
 */
static long
asn_date (char *date_string)
{
  time_t      current_time;
  struct tm  *bd_time;
#ifndef HAVE_TIMEZONE
  #ifdef TM_GM_TOFF
     #define timezone (bdtime->gm_toff)
  #else
     #define timezone 0l
  #endif /* TM_GM_TOFF */
#endif /* HAVE_TIMEZONE */

  time(&current_time);
  bd_time = localtime(&current_time);
  sprintf(date_string, "D:%04d%02d%02d%02d%02d%02d%+03ld'%02ld'",
	  bd_time->tm_year + 1900, bd_time->tm_mon + 1, bd_time->tm_mday,
	  bd_time->tm_hour, bd_time->tm_min, bd_time->tm_sec,
	  -timezone/3600, timezone % 3600);

  return strlen(date_string);
}

#if 0
/*
 * Placing copyright notice of the software in Producer string
 * is inappropriate.
 */
#define BANNER "\
 Copyright \251 2002-2004  by Jin-Hwan Cho and Shunsaku Hirata,\
 Copyright \251 1998, 1999 by Mark A. Wicks.\
"
#endif

static void
doc_close_docinfo (void)
{
  /*
   * Excerpt from PDF Reference 4th ed., sec. 10.2.1.
   *
   * Any entry whose value is not known should be omitted from the dictionary,
   * rather than included with an empty string as its value.
   *
   * ....
   *
   * Note: Although viewer applications can store custom metadata in the document
   * information dictionary, it is inappropriate to store private content or
   * structural information there; such information should be stored in the
   * document catalog instead (see Section 3.6.1,  Document Catalog ).
   */
  const char *keys[] = {
    "Title", "Author", "Subject", "Keywords", "Creator", "Producer",
    "CreationDate", "ModDate", /* Date */
    NULL
  };
  pdf_obj *value;
  char    *banner;
  int      i;

  for (i = 0; keys[i] != NULL; i++) {
    value = pdf_lookup_dict(docinfo, keys[i]);
    if (value) {
      if (!PDF_OBJ_STRINGTYPE(value)) {
	WARN("\"%s\" in DocInfo dictionary not string type.", keys[i]);
	pdf_remove_dict(docinfo, keys[i]);
	WARN("\"%s\" removed from DocInfo.", keys[i]);
      } else if (pdf_string_length(value) == 0) {
	/*
	 * The hyperref package often uses emtpy strings.
	 */
	pdf_remove_dict(docinfo, keys[i]);
      }
    }
  }

#if 0
  /* See above */
  banner = NEW(strlen(BANNER)+strlen(PACKAGE)+strlen(VERSION)+16, char);
  sprintf(banner, "%s (%s), %s", PACKAGE, VERSION, BANNER);
#endif
  banner = NEW(strlen(PACKAGE)+strlen(VERSION)+4, char);
  sprintf(banner, "%s (%s)", PACKAGE, VERSION);
  pdf_add_dict(docinfo,
	       pdf_new_name("Producer"),
	       pdf_new_string(banner, strlen(banner)));
  RELEASE(banner);
  
  if (!pdf_lookup_dict(docinfo, "CreationDate")) {
    char now[32];

    asn_date(now);
    pdf_add_dict(docinfo, 
		 pdf_new_name ("CreationDate"),
		 pdf_new_string(now, strlen(now)));
  }

  pdf_release_obj(docinfo);
}

static pdf_obj *
get_currentpage_resources (const char *category)
{
  pdf_obj *resources;
  struct page_entry *currentpage;

  if (!category) {
    return NULL;
  }
  currentpage = LASTPAGE();
  if (!currentpage->resources) {
    currentpage->resources = pdf_new_dict();
  }
  resources = pdf_lookup_dict(currentpage->resources, category);
  if (!resources) {
    resources = pdf_new_dict();
    pdf_add_dict(currentpage->resources, pdf_new_name(category), resources);
  }

  return resources;
}

void
pdf_doc_add_page_resource (const char *category,
			   const char *resource_name, pdf_obj *resource_ref)
{
  pdf_obj *resources;
  pdf_obj *duplicate;

  if (!PDF_OBJ_INDIRECTTYPE(resource_ref)) {
    WARN("Passed non indirect reference...");
    resource_ref = pdf_ref_obj(resource_ref); /* leak */
  }
  resources = get_currentpage_resources(category);
  duplicate = pdf_lookup_dict(resources, resource_name);
  if (duplicate && pdf_compare_reference(duplicate, resource_ref)) {
    WARN("Conflicting page resource found (page: %ld, category: %s, name: %s).",
	 pdf_doc_current_page_number(), category, resource_name);
    WARN("Ignoring...");
    pdf_release_obj(resource_ref);
  } else {
    pdf_add_dict(resources, pdf_new_name(resource_name), resource_ref);
  }

  return;
}

static void
doc_flush_page (struct page_entry *page, pdf_obj *parent_ref)
{
  pdf_obj *contents_array;
  int      count;

  pdf_add_dict(page->page_obj,
	       pdf_new_name("Type"), pdf_new_name("Page"));
  pdf_add_dict(page->page_obj,
	       pdf_new_name("Parent"), parent_ref);

  /*
   * Clipping area specified by CropBox is affected by MediaBox which
   * might be inherit from parent node. If MediaBox of the root node
   * does not have enough size to cover all page's imaging area, using
   * CropBox here gives incorrect result.
   */
  if (page->flags & USE_MY_MEDIABOX) {
    pdf_obj *mediabox;

    mediabox = pdf_new_array();
    pdf_add_array(mediabox,
		  pdf_new_number(ROUND(page->cropbox.llx, 0.01)));
    pdf_add_array(mediabox,
		  pdf_new_number(ROUND(page->cropbox.lly, 0.01)));
    pdf_add_array(mediabox,
		  pdf_new_number(ROUND(page->cropbox.urx, 0.01)));
    pdf_add_array(mediabox,
		  pdf_new_number(ROUND(page->cropbox.ury, 0.01)));
    pdf_add_dict(page->page_obj, pdf_new_name("MediaBox"),  mediabox);
  }

  count = 0;
  contents_array = pdf_new_array();
  if (page->content_refs[0]) { /* global bop */
    pdf_add_array(contents_array, page->content_refs[0]);
    count++;
  } else if (docpages.bop &&
	     pdf_stream_length(docpages.bop) > 0) {
    pdf_add_array(contents_array, pdf_ref_obj(docpages.bop));
    count++;
  }
  if (page->content_refs[1]) { /* background */
    pdf_add_array(contents_array, page->content_refs[1]);
    count++;
  }
  if (page->content_refs[2]) { /* page body */
    pdf_add_array(contents_array, page->content_refs[2]);
    count++;
  }
  if (page->content_refs[3]) { /* global eop */
    pdf_add_array(contents_array, page->content_refs[3]);
    count++;
  } else if (docpages.eop &&
	     pdf_stream_length(docpages.eop) > 0) {
    pdf_add_array(contents_array, pdf_ref_obj(docpages.eop));
    count++;
  }

  if (count == 0)
    WARN("Page with empty content found!!!");
  page->content_refs[0] = NULL;
  page->content_refs[1] = NULL;
  page->content_refs[2] = NULL;
  page->content_refs[3] = NULL;

  pdf_add_dict(page->page_obj,
	       pdf_new_name("Contents"), contents_array);


  if (page->annots) {
    pdf_add_dict(page->page_obj,
		 pdf_new_name("Annots"), pdf_ref_obj(page->annots));
    pdf_release_obj(page->annots);
  }
  if (page->beads) {
    pdf_add_dict(page->page_obj,
		 pdf_new_name("B"), pdf_ref_obj(page->beads));
    pdf_release_obj(page->beads);
  }
  pdf_release_obj(page->page_obj);
  pdf_release_obj(page->page_ref);

  page->page_obj = NULL;
  page->page_ref = NULL;
  page->annots   = NULL;
  page->beads    = NULL;
}

/* B-tree? */
#define PAGE_CLUSTER 4
static pdf_obj *
build_page_tree (struct page_entry *firstpage,
		 long num_pages, pdf_obj *parent_ref)
{
  pdf_obj *self, *self_ref, *kids;
  long i;

  self = pdf_new_dict();
  /*
   * This is a slight kludge which allow the subtree dictionary
   * generated by this routine to be merged with the real
   * page_tree dictionary, while keeping the indirect object
   * references right.
   */
  self_ref = parent_ref ? pdf_ref_obj(self) : pdf_ref_obj(docroot.pages);

  pdf_add_dict(self, pdf_new_name("Type"),  pdf_new_name("Pages"));
  pdf_add_dict(self, pdf_new_name("Count"), pdf_new_number((double) num_pages));

  if (parent_ref != NULL)
    pdf_add_dict(self, pdf_new_name("Parent"), parent_ref);

  kids = pdf_new_array();
  if (num_pages > 0 && num_pages <= PAGE_CLUSTER) {
    for (i = 0; i < num_pages; i++) {
      struct page_entry *page;

      page = firstpage + i;
      if (!page->page_ref)
	page->page_ref = pdf_ref_obj(page->page_obj);
      pdf_add_array(kids, pdf_link_obj(page->page_ref));
      doc_flush_page(page, pdf_link_obj(self_ref));
    }
  } else if (num_pages > 0) {
    for (i = 0; i < PAGE_CLUSTER; i++) {
      long start, end;

      start = (i*num_pages)/PAGE_CLUSTER;
      end   = ((i+1)*num_pages)/PAGE_CLUSTER;
      if (end - start > 1) {
	pdf_obj *subtree;

	subtree = build_page_tree(firstpage + start, end - start, pdf_link_obj(self_ref));
	pdf_add_array(kids, pdf_ref_obj(subtree));
	pdf_release_obj(subtree);
      } else {
	struct page_entry *page;

	page = firstpage + start;
	if (!page->page_ref)
	  page->page_ref = pdf_ref_obj(page->page_obj);
	pdf_add_array(kids, pdf_link_obj(page->page_ref));
	doc_flush_page(page, pdf_link_obj(self_ref));
      }
    }
  }
  pdf_add_dict(self, pdf_new_name("Kids"), kids);
  pdf_release_obj(self_ref);

  return self;
}

static void
doc_init_page_tree (double media_width, double media_height)
{
  /*
   * Create empty page tree.
   * The docroot.pages is kept open until the document is closed.
   * This allows the user to write to pages if he so choses.
   */
  docroot.pages = pdf_new_dict();

  docpages.bop = NULL;
  docpages.eop = NULL;

  docpages.num_pages = 0;
  docpages.max_pages = 0;
  docpages.pages     = NULL;

  docpages.mediabox.llx = 0.0;
  docpages.mediabox.lly = 0.0;
  docpages.mediabox.urx = media_width;
  docpages.mediabox.ury = media_height;
}

static void
doc_close_page_tree (void)
{
  pdf_obj *page_tree_root;
  pdf_obj *mediabox;
  long     page_no;

  /*
   * Do consistency check on forward references to pages.
   */
  for (page_no = PAGECOUNT() + 1; page_no <= MAXPAGES(); page_no++) {
    struct page_entry *page;

    page = doc_page_entry(page_no);
    if (page->page_obj) {
      WARN("Nonexistent page #%ld refered.", page_no);
      pdf_release_obj(page->page_ref);
      page->page_ref = NULL;
    }
    if (page->page_obj) {
      WARN("Entry for a nonexistent page #%ld created.", page_no);
      pdf_release_obj(page->page_obj);
      page->page_obj = NULL;
    }
    if (page->annots) {
      WARN("Annotation attached to a nonexistent page #%ld.", page_no);
      pdf_release_obj(page->annots);
      page->annots = NULL;
    }
    if (page->beads) {
      WARN("Article beads attached to a nonexistent page #%ld.", page_no);
      pdf_release_obj(page->beads);
      page->beads = NULL;
    }
    if (page->resources) {
      pdf_release_obj(page->resources);
      page->resources = NULL;
    }
  }

  /*
   * Connect page tree to root node.
   */
  page_tree_root = build_page_tree(FIRSTPAGE(), PAGECOUNT(), NULL);
  pdf_merge_dict(docroot.pages, page_tree_root);
  pdf_release_obj(page_tree_root);

  /* They must be after build_page_tree() */
  if (docpages.bop) {
    pdf_add_stream (docpages.bop, "\n", 1);
    pdf_release_obj(docpages.bop);
    docpages.bop = NULL;
  }
  if (docpages.eop) {
    pdf_add_stream (docpages.eop, "\n", 1);
    pdf_release_obj(docpages.eop);
    docpages.eop = NULL;
  }

  /* Create media box at root node and let the other pages inherit it. */
  mediabox = pdf_new_array();
  pdf_add_array(mediabox, pdf_new_number(ROUND(docpages.mediabox.llx, 0.01)));
  pdf_add_array(mediabox, pdf_new_number(ROUND(docpages.mediabox.lly, 0.01)));
  pdf_add_array(mediabox, pdf_new_number(ROUND(docpages.mediabox.urx, 0.01)));
  pdf_add_array(mediabox, pdf_new_number(ROUND(docpages.mediabox.ury, 0.01)));
  pdf_add_dict(docroot.pages, pdf_new_name("MediaBox"), mediabox);

  pdf_add_dict(docroot.catalog,
	       pdf_new_name("Pages"),
	       pdf_ref_obj(docroot.pages));
  pdf_release_obj(docroot.pages);
  docroot.pages  = NULL;

  RELEASE(docpages.pages);
  docpages.pages     = NULL;
  docpages.num_pages = 0;
  docpages.max_pages = 0;
}

/*
 * Outline (Bookmark)
 *
 * This doesn't remember outline hierarcy at all...
 */

#ifndef BOOKMARKS_OPEN_DEFAULT
#define BOOKMARKS_OPEN_DEFAULT 0
#endif
static int outline_depth      = 0;
static int outline_open_depth = 0;

#define MAX_OUTLINE_DEPTH 32u

static struct bookmark_item
{
  int      count;
  int      is_open;
  pdf_obj *object;
} outline[MAX_OUTLINE_DEPTH];

static void
doc_init_bookmarks (int bookmark_open_depth)
{
  outline_open_depth = ((bookmark_open_depth >= 0) ?
			bookmark_open_depth : MAX_OUTLINE_DEPTH - bookmark_open_depth);
  outline_depth = 0;
  /* Create empty outline tree */
  outline[outline_depth].object  = pdf_new_dict();
  outline[outline_depth].count   = 0;
  outline[outline_depth].is_open = 0;
}

int
pdf_doc_bookmarks_up (void)
{
  int count;

  if (outline_depth == 0)
    ERROR("Can't go up above the root node!");

  /*
   * Terminate lower level entries.
   */
  pdf_add_dict(outline[outline_depth-1].object,
	       pdf_new_name("Last"), pdf_ref_obj(outline[outline_depth].object));
  pdf_release_obj(outline[outline_depth].object);
  outline[outline_depth].object  = NULL;
  outline[outline_depth].count   = 0;
  outline[outline_depth].is_open = 0;

  /*
   * If the item is open, the value of Count is a positive number which
   * represents the total number of its open descendants at all lower
   * levels of the outline hierarchy. If the item is closed, it is a
   * negative integer whose absolute value specifies how many descendants
   * would appear if the item were reopened.
   */
  if (outline_depth > 1) {
    if (outline[outline_depth-1].is_open) {
      count = outline[outline_depth-1].count;
      /*
       * The count of parent node is a number of it's immediate descendant
       * node at this time. Add all open descendants at all lower levels.
       */
      outline[outline_depth-2].count += count;
    } else {
      count = -outline[outline_depth-1].count;
    }
  } else {
    /* Top level. Can't close it. */
    count = outline[0].count;
  }
  pdf_add_dict(outline[outline_depth-1].object,
	       pdf_new_name("Count"), pdf_new_number(count));

  outline_depth--;

  return outline_depth;
}

int
pdf_doc_bookmarks_down (void)
{
  if (outline_depth >= MAX_OUTLINE_DEPTH - 1)
    ERROR("Outline is too deep.");

  if (!outline[outline_depth].object) {
    pdf_obj *bookmark_item, *text_color, *action;

    WARN("Empty bookmark node!");
    WARN("You have tried to jump more than 1 level.");

#define TITLE_STRING "<No Title>"
    bookmark_item = pdf_new_dict();
    pdf_add_dict(bookmark_item,
		 pdf_new_name("Title"),
		 pdf_new_string(TITLE_STRING, strlen(TITLE_STRING)));

    text_color = pdf_new_array();
    pdf_add_array(text_color, pdf_new_number(1.0));
    pdf_add_array(text_color, pdf_new_number(0.0));
    pdf_add_array(text_color, pdf_new_number(0.0));
    pdf_add_dict(bookmark_item, pdf_new_name("C"), text_color);

    pdf_add_dict(bookmark_item, pdf_new_name("F"), pdf_new_number(1.0));

#define JS_CODE "app.alert(\"The author of this document made this bookmark item empty!\", 3, 0)"
    action = pdf_new_dict();
    pdf_add_dict(action, pdf_new_name("S"), pdf_new_name("JavaScript"));
    pdf_add_dict(action, 
		 pdf_new_name("JS"), pdf_new_string(JS_CODE, strlen(JS_CODE)));
    pdf_add_dict(bookmark_item, pdf_new_name("A"), action);

    pdf_doc_bookmarks_add(bookmark_item, 0);
  }

  outline_depth++;

  return outline_depth;
}

int
pdf_doc_bookmarks_depth (void)
{
  return outline_depth;
}

void
pdf_doc_bookmarks_add (pdf_obj *dict, int is_open)
{
  pdf_obj *new_item;

  if (outline_depth < 1)
    ERROR("Can't add to outline at depth < 1.");

  if (is_open < 0) {
    if (outline_depth <= outline_open_depth)
      is_open = 1;
    else
      is_open = 0;
  }

  new_item = pdf_new_dict();
  /*
   * Caller doesn't know we don't actually use the dictionary,
   * so he *gave* dict to us.  We have to free it.
   */
  pdf_merge_dict (new_item, dict);
  pdf_release_obj(dict);

  pdf_add_dict(new_item,
	       pdf_new_name("Parent"),
	       pdf_ref_obj(outline[outline_depth-1].object));
  /*
   * We don't care about open items at lower levels at this moment.
   */
  outline[outline_depth-1].count += 1;

  if (outline[outline_depth].object == NULL) {
    pdf_add_dict(outline[outline_depth-1].object,
		 pdf_new_name("First"), pdf_ref_obj(new_item));
  } else {
    pdf_add_dict(new_item,
		 pdf_new_name("Prev"),
		 pdf_ref_obj(outline[outline_depth].object));
    pdf_add_dict(outline[outline_depth].object,
		 pdf_new_name("Next"), pdf_ref_obj(new_item));
    pdf_release_obj(outline[outline_depth].object);
  }
  outline[outline_depth].object  = new_item;
  outline[outline_depth].count   = 0;
  outline[outline_depth].is_open = is_open;
}

static void
doc_close_bookmarks (void)
{
  if (outline[0].object) {
    while (outline_depth > 0) {
      /* pdf_doc_bookmarks_up() decreases outline_depth. */
      pdf_doc_bookmarks_up();
    }
    if (outline[0].count > 0) {
      pdf_add_dict(docroot.catalog,
		   pdf_new_name("Outlines"), pdf_ref_obj(outline[0].object));
    }
    pdf_release_obj(outline[0].object);
    outline[0].object = NULL;
  }
  outline_depth = 0;
}

#define NUM_NAME_CATEGORY 10
static struct
{
  const char      *category;
  struct ht_table *data;
} docnames[] = {
  {"Dests",      NULL},
  {"AP",         NULL},
  {"JavaScript", NULL},
  {"Pages",      NULL},
  {"Templates",  NULL},
  {"IDS",        NULL},
  {"URLS",       NULL},
  {"EmbeddedFiles", NULL},
  {"AlternatePresentations", NULL},
  {"Renditions", NULL},
  {NULL, NULL}
};

static void
doc_init_names (void)
{
  int i;

  docroot.names   = NULL;
  for (i = 0; docnames[i].category != NULL; i++) {
    docnames[i].data = NULL;
  }
}

int
pdf_doc_add_names (const char *category,
		   const void *key, int keylen, pdf_obj *value)
{
  int   i;

  for (i = 0; docnames[i].category != NULL; i++) {
    if (!strcmp(docnames[i].category, category)) {
      break;
    }
  }
  if (docnames[i].category == NULL) {
    WARN("Unknown name dictionary category \"%s\".", category);
    return -1;
  }
  if (!docnames[i].data) {
    docnames[i].data = pdf_new_name_tree();
  }

  return pdf_names_add_object(docnames[i].data, key, keylen, value);
}

static void
doc_close_names (void)
{
  pdf_obj  *tmp;
  int       i;

  for (i = 0; docnames[i].category != NULL; i++) {
    if (docnames[i].data) {
      pdf_obj *name_tree;

      name_tree = pdf_names_create_tree(docnames[i].data);
      if (!docroot.names) {
	docroot.names = pdf_new_dict();
      }
      pdf_add_dict(docroot.names,
		   pdf_new_name(docnames[i].category), pdf_ref_obj(name_tree));
      pdf_release_obj(name_tree);
      pdf_delete_name_tree(&docnames[i].data);
    }
  }

  if (docroot.names) {
    tmp = pdf_lookup_dict(docroot.catalog, "Names");
    if (!tmp) {
      pdf_add_dict(docroot.catalog,
		   pdf_new_name("Names"), pdf_ref_obj(docroot.names));
    } else if (PDF_OBJ_DICTTYPE(tmp)) {
      pdf_merge_dict(docroot.names, tmp);
      pdf_add_dict(docroot.catalog,
		   pdf_new_name("Names"), pdf_ref_obj(docroot.names));
    } else { /* Maybe reference */
      /* What should I do? */
      WARN("Could not modify Names dictionary.");
    }
    pdf_release_obj(docroot.names);
    docroot.names = NULL;
  }
}


static double annot_grow = 0.0; /* ... */

void
pdf_doc_add_annot (unsigned page_no, const pdf_rect *rect, pdf_obj *annot_dict)
{
  struct page_entry *page;
  pdf_obj  *rect_array;
  pdf_rect  mediabox;

  page = doc_page_entry(page_no);
  if (!page->annots)
    page->annots = pdf_new_array();

  pdf_doc_get_mediabox(pdf_doc_current_page_number(), &mediabox);
  if (rect->llx < mediabox.llx ||
      rect->urx > mediabox.urx ||
      rect->lly < mediabox.lly ||
      rect->ury > mediabox.ury) {
    WARN("Annotation out of page boundary.");
    WARN("Current page's MediaBox: [%g %g %g %g]",
	 mediabox.llx, mediabox.lly, mediabox.urx, mediabox.ury);
    WARN("Annotation: [%g %g %g %g]",
	 rect->llx, rect->lly, rect->urx, rect->ury);
    WARN("Maybe incorrect paper size specified.");
  }
  if (rect->llx > rect->urx || rect->lly > rect->ury) {
    WARN("Rectangle with negative width/height: [%g %g %g %g]",
	 rect->llx, rect->lly, rect->urx, rect->ury);
  }

  rect_array = pdf_new_array();
  pdf_add_array(rect_array, pdf_new_number(ROUND(rect->llx - annot_grow, 0.01)));
  pdf_add_array(rect_array, pdf_new_number(ROUND(rect->lly - annot_grow, 0.01)));
  pdf_add_array(rect_array, pdf_new_number(ROUND(rect->urx + annot_grow, 0.01)));
  pdf_add_array(rect_array, pdf_new_number(ROUND(rect->ury + annot_grow, 0.01)));
  pdf_add_dict (annot_dict, pdf_new_name("Rect"), rect_array);

  pdf_add_array(page->annots, pdf_ref_obj(annot_dict));
}

/*
 * PDF Article Thread
 */
struct article_bead
{
  char    *id;
  long     page_no;
  pdf_rect rect;
};

struct article_thread
{
  char    *id;
  pdf_obj *info;
  long num_beads;
  long max_beads;
  struct article_bead *beads;
};

static struct {
  long num_articles;
  long max_articles;
  struct article_thread *articles;
} docarticles = {0, 0, NULL};

static void
doc_init_articles (void)
{
  docroot.threads = NULL;

  docarticles.num_articles = 0;
  docarticles.max_articles = 0;
  docarticles.articles     = NULL;
}

void
pdf_doc_begin_article (const char *article_id, pdf_obj *article_info)
{
  struct article_thread *article;

  if (article_id == NULL || strlen(article_id) == 0)
    ERROR("Article thread without internal identifier.");

  if (docarticles.num_articles >= docarticles.max_articles) {
    docarticles.max_articles += PDFDOC_ARTICLE_ALLOC_SIZE;
    docarticles.articles = RENEW(docarticles.articles,
				 docarticles.max_articles, struct article_thread);
  }
  article = &(docarticles.articles[docarticles.num_articles]);

  article->id = NEW(strlen(article_id)+1, char);
  strcpy(article->id, article_id);
  article->info = article_info;
  article->num_beads = 0;
  article->max_beads = 0;
  article->beads     = NULL;

  docarticles.num_articles++;
}

void
pdf_doc_end_article (const char *article_id)
{
  return; /* no-op */
}

static struct article_bead *
find_bead (struct article_thread *article, const char *bead_id)
{
  struct article_bead *bead;
  long i;

  bead = NULL;
  for (i = 0; i < article->num_beads; i++) {
    if (!strcmp(article->beads[i].id, bead_id)) {
      bead = &(article->beads[i]);
      break;
    }
  }

  return bead;
}

void
pdf_doc_add_bead (const char *article_id,
		  const char *bead_id, long page_no, const pdf_rect *rect)
{
  struct article_thread *article;
  struct article_bead   *bead;
  long i;

  if (!article_id) {
    ERROR("No article identifier specified.");
  }

  article = NULL;
  for (i = 0; i < docarticles.num_articles; i++) {
    if (!strcmp(docarticles.articles[i].id, article_id)) {
      article = &(docarticles.articles[i]);
      break;
    }
  }
  if (!article) {
    ERROR("Specified article thread that doesn't exist.");
    return;
  }

  bead = bead_id ? find_bead(article, bead_id) : NULL;
  if (!bead) {
    if (article->num_beads >= article->max_beads) {
      article->max_beads += PDFDOC_BEAD_ALLOC_SIZE;
      article->beads = RENEW(article->beads,
			     article->max_beads, struct article_bead);
      for (i = article->num_beads; i < article->max_beads; i++) {
	article->beads[i].id = NULL;
	article->beads[i].page_no = -1;
      }
    }
    bead = &(article->beads[article->num_beads]);
    if (bead_id) {
      bead->id = NEW(strlen(bead_id)+1, char);
      strcpy(bead->id, bead_id);
    } else {
      bead->id = NULL;
    }
    article->num_beads++;
  }
  bead->rect.llx = rect->llx;
  bead->rect.lly = rect->lly;
  bead->rect.urx = rect->urx;
  bead->rect.ury = rect->ury;
  bead->page_no  = page_no;
}

static pdf_obj *
doc_make_article (struct article_thread *article,
		  const char **bead_ids, int num_beads,
		  pdf_obj *article_info)
{
  pdf_obj *art_dict;
  pdf_obj *first, *prev, *last;
  long i, n;

  if (!article)
    return NULL;

  art_dict = pdf_new_dict();
  first = prev = last = NULL;
  /*
   * The bead_ids represents logical order of beads in an article thread.
   * If bead_ids is not given, we create an article thread in the order of
   * beads appeared.
   */
  n = bead_ids ? num_beads : article->num_beads;
  for (i = 0; i < n; i++) {
    struct article_bead *bead;

    bead = bead_ids ? find_bead(article, bead_ids[i]) : &(article->beads[i]);
    if (!bead || bead->page_no < 0) {
      continue;
    }
    last = pdf_new_dict();
    if (prev == NULL) {
      first = last;
      pdf_add_dict(first,
		   pdf_new_name("T"), pdf_ref_obj(art_dict));
    } else {
      pdf_add_dict(prev,
		   pdf_new_name("N"), pdf_ref_obj(last));
      pdf_add_dict(last,
		   pdf_new_name("V"), pdf_ref_obj(prev));
      /* We must link first to last. */
      if (prev != first)
	pdf_release_obj(prev);
    }

    /* Realize bead now. */
    {
      struct page_entry *page;
      pdf_obj *rect;

      page = doc_page_entry(bead->page_no);
      if (page->beads == NULL) {
	page->beads = pdf_new_array();
      }
      pdf_add_dict(last, pdf_new_name("P"), pdf_link_obj(page->page_ref));
      rect = pdf_new_array();
      pdf_add_array(rect, pdf_new_number(ROUND(bead->rect.llx, 0.01)));
      pdf_add_array(rect, pdf_new_number(ROUND(bead->rect.lly, 0.01)));
      pdf_add_array(rect, pdf_new_number(ROUND(bead->rect.urx, 0.01)));
      pdf_add_array(rect, pdf_new_number(ROUND(bead->rect.ury, 0.01)));
      pdf_add_dict (last, pdf_new_name("R"), rect);
      pdf_add_array(page->beads, pdf_ref_obj(last));
    }

    prev = last;
  }

  if (first && last) {
    pdf_add_dict(last,
		 pdf_new_name("N"), pdf_ref_obj(first));
    pdf_add_dict(first,
		 pdf_new_name("V"), pdf_ref_obj(last));
    pdf_release_obj(last);
    pdf_add_dict(art_dict,
		 pdf_new_name("F"), pdf_ref_obj(first));
    /* If article_info is supplied, we override article->info. */
    if (article_info) {
      pdf_add_dict(art_dict,
		   pdf_new_name("I"), article_info);
    } else if (article->info) {
      pdf_add_dict(art_dict,
		   pdf_new_name("I"), article->info);
      article->info = NULL; /* We do not write as object reference. */
    }
    pdf_release_obj(first);
  } else {
    pdf_release_obj(art_dict);
    art_dict = NULL;
  }

  return art_dict;
}

#if 0
void
pdf_doc_add_article (const char *article_id,
		     const char **bead_ids, int num_beads,
		     pdf_obj *article_info)
{
  long i;

  if (!article_id)
    return;

  for (i = 0; i < docarticles.num_articles; i++) {
    struct article_thread *article;

    article = &(docarticles.articles[i]);
    if (article->id && !strcmp(article->id, article_id)) {
      pdf_obj *art_dict;

      art_dict = doc_make_article(article, bead_ids, num_beads, article_info);
      if (!docroot.threads) {
	docroot.threads = pdf_new_array();
      }
      pdf_add_array(docroot.threads, pdf_ref_obj(art_dict));
      pdf_release_obj(art_dict);

      break;
    }
  }
}
#endif

static void
doc_clear_article (struct article_thread *article)
{
  if (article) {
    if (article->beads) {
      long i;
      for (i = 0; i < article->num_beads; i++) {
	if (article->beads[i].id)
	  RELEASE(article->beads[i].id);
      }
      RELEASE(article->beads);
      article->beads = NULL;
    }
    if (article->id)
      RELEASE(article->id);
    article->id = NULL;
    article->num_beads = 0;
    article->max_beads = 0;
  }
}

static void
doc_close_articles (void)
{
  int i;

  for (i = 0; i < docarticles.num_articles; i++) {
    struct article_thread *article;

    article = &(docarticles.articles[i]);
    if (article->beads) {
      pdf_obj *art_dict;

      art_dict = doc_make_article(article, NULL, 0, NULL);
      if (!docroot.threads) {
	docroot.threads = pdf_new_array();
      }
      pdf_add_array(docroot.threads, pdf_ref_obj(art_dict));
      pdf_release_obj(art_dict);
    }
    doc_clear_article(article);
  }
  RELEASE(docarticles.articles);
  docarticles.articles = NULL;
  docarticles.num_articles = 0;
  docarticles.max_articles = 0;

  if (docroot.threads) {
    pdf_add_dict(docroot.catalog,
		 pdf_new_name("Threads"), pdf_ref_obj(docroot.threads));
    pdf_release_obj(docroot.threads);
    docroot.threads = NULL;
  }
}

static int xobjects_pending = 0;

/* page_no = 0 for root page tree node. */
void
pdf_doc_set_mediabox (unsigned page_no, const pdf_rect *mediabox)
{
  struct page_entry *page;

  if (page_no == 0) {
    docpages.mediabox.llx = mediabox->llx;
    docpages.mediabox.lly = mediabox->lly;
    docpages.mediabox.urx = mediabox->urx;
    docpages.mediabox.ury = mediabox->ury;
  } else {
    page = doc_page_entry(page_no);
    page->cropbox.llx = mediabox->llx;
    page->cropbox.lly = mediabox->lly;
    page->cropbox.urx = mediabox->urx;
    page->cropbox.ury = mediabox->ury;
    page->flags |= USE_MY_MEDIABOX;
  }
}

void
pdf_doc_get_mediabox (unsigned page_no, pdf_rect *mediabox)
{
  struct page_entry *page;

  if (page_no == 0) {
    mediabox->llx = docpages.mediabox.llx;
    mediabox->lly = docpages.mediabox.lly;
    mediabox->urx = docpages.mediabox.urx;
    mediabox->ury = docpages.mediabox.ury;
  } else {
    page = doc_page_entry(page_no);
    if (page->flags & USE_MY_MEDIABOX) {
      mediabox->llx = page->cropbox.llx;
      mediabox->lly = page->cropbox.lly;
      mediabox->urx = page->cropbox.urx;
      mediabox->ury = page->cropbox.ury;
    } else {
      mediabox->llx = docpages.mediabox.llx;
      mediabox->lly = docpages.mediabox.lly;
      mediabox->urx = docpages.mediabox.urx;
      mediabox->ury = docpages.mediabox.ury;
    }
  }
}

pdf_obj *
pdf_doc_current_page_resources (void)
{
  struct page_entry *currentpage;

  currentpage = LASTPAGE();

  if (!currentpage->resources)
    currentpage->resources = pdf_new_dict();

  return currentpage->resources;
}

pdf_obj *
pdf_doc_get_dictionary (const char *category)
{
  pdf_obj *dict = NULL;

  ASSERT(category);

  if (!strcmp(category, "Names")) {
    if (!docroot.names)
      docroot.names = pdf_new_dict();
    dict = docroot.names;
  } else if (!strcmp(category, "Pages")) {
    if (!docroot.pages)
      docroot.pages = pdf_new_dict();
    dict = docroot.pages;
  } else if (!strcmp(category, "Catalog")) {
    if (!docroot.catalog)
      docroot.catalog = pdf_new_dict();
    dict = docroot.catalog;
  } else if (!strcmp(category, "Info")) {
    if (!docinfo)
      docinfo = pdf_new_dict();
    dict = docinfo;
  } else if (!strcmp(category, "@THISPAGE")) {
    /* Sorry for this... */
    struct page_entry *currentpage;

    currentpage = LASTPAGE();
    dict =  currentpage->page_obj;
  }

  if (!dict) {
    ERROR("Document dict. \"%s\" not exist. ", category);
  }

  return dict;
}

long
pdf_doc_current_page_number (void)
{
  return (long) (PAGECOUNT() + 1);
}

pdf_obj *
pdf_doc_ref_page (unsigned long page_no)
{
  struct page_entry *page;

  page = doc_page_entry(page_no);
  if (!page->page_obj) {
    page->page_obj = pdf_new_dict();
    page->page_ref = pdf_ref_obj(page->page_obj);
  }

  return pdf_link_obj(page->page_ref);
}

pdf_obj *
pdf_doc_get_reference (const char *category)
{
  pdf_obj *ref = NULL;
  long     page_no;

  ASSERT(category);

  page_no = pdf_doc_current_page_number();
  if (!strcmp(category, "@THISPAGE")) {
    ref = pdf_doc_ref_page(page_no);
  } else if (!strcmp(category, "@PREVPAGE")) {
    if (page_no <= 1) {
      ERROR("Reference to previous page, but no pages have been completed yet.");
    }
    ref = pdf_doc_ref_page(page_no - 1);
  } else if (!strcmp(category, "@NEXTPAGE")) {
    ref = pdf_doc_ref_page(page_no + 1);
  }

  if (!ref) {
    ERROR("Reference to \"%s\" not exist. ", category);
  }

  return ref;
}

static void
pdf_doc_new_page (void)
{
  struct page_entry *currentpage;

  if (PAGECOUNT() >= MAXPAGES()) {
    resize_pages(MAXPAGES() + PDFDOC_PAGES_ALLOC_SIZE);
  }

  /*
   * This is confusing. pdf_doc_finish_page() have increased page count!
   */
  currentpage = LASTPAGE();
  /* Was this page already instantiated by a forward reference to it? */
  if (!currentpage->page_ref) {
    currentpage->page_obj = pdf_new_dict();
    currentpage->page_ref = pdf_ref_obj(currentpage->page_obj);
  }

  currentpage->background = NULL;
  currentpage->contents   = pdf_new_stream(STREAM_COMPRESS);
  currentpage->resources  = pdf_new_dict();

  currentpage->annots = NULL;
  currentpage->beads  = NULL;
}

/* This only closes contents and resources. */
static void
pdf_doc_finish_page (void)
{
  struct page_entry *currentpage;

  if (xobjects_pending)
    ERROR("A pending form XObject at end of page.");

  currentpage = LASTPAGE();
  if (!currentpage->page_obj)
    currentpage->page_obj = pdf_new_dict();

  /*
   * Make Contents array.
   */

  /*
   * Global BOP content stream.
   * pdf_ref_obj() returns reference itself when the object is
   * indirect reference, not reference to the indirect reference.
   * We keep bop itself but not reference to it since it is
   * expected to be small.
   */
  if (docpages.bop &&
      pdf_stream_length(docpages.bop) > 0) {
    currentpage->content_refs[0] = pdf_ref_obj(docpages.bop);
  } else {
    currentpage->content_refs[0] = NULL;
  }
  /*
   * Current page background content stream.
   */
  if (currentpage->background) {
    if (pdf_stream_length(currentpage->background) > 0) {
      currentpage->content_refs[1] = pdf_ref_obj(currentpage->background);
      pdf_add_stream (currentpage->background, "\n", 1);
    }
    pdf_release_obj(currentpage->background);
    currentpage->background = NULL;
  } else {
    currentpage->content_refs[1] = NULL;
  }

  /* Content body of current page */
  currentpage->content_refs[2] = pdf_ref_obj(currentpage->contents);
  pdf_add_stream (currentpage->contents, "\n", 1);
  pdf_release_obj(currentpage->contents);
  currentpage->contents = NULL;

  /*
   * Global EOP content stream.
   */
  if (docpages.eop &&
      pdf_stream_length(docpages.eop) > 0) {
    currentpage->content_refs[3] = pdf_ref_obj(docpages.eop);
  } else {
    currentpage->content_refs[3] = NULL;
  }

  /*
   * Page resources.
   */
  if (currentpage->resources) {
    pdf_obj *procset;
    /*
     * ProcSet is obsolete in PDF-1.4 but recommended for compatibility.
     */

    procset = pdf_new_array ();
    pdf_add_array(procset, pdf_new_name("PDF"));
    pdf_add_array(procset, pdf_new_name("Text"));
    pdf_add_array(procset, pdf_new_name("ImageC"));
    pdf_add_array(procset, pdf_new_name("ImageB"));
    pdf_add_array(procset, pdf_new_name("ImageI"));
    pdf_add_dict(currentpage->resources, pdf_new_name("ProcSet"), procset);

    pdf_add_dict(currentpage->page_obj,
		 pdf_new_name("Resources"),
		 pdf_ref_obj(currentpage->resources));
    pdf_release_obj(currentpage->resources);
    currentpage->resources = NULL;
  }

#ifndef NO_THUMBNAIL
  if (thumb_enabled) {
    char    *thumb_filename;
    pdf_obj *thumb_ref;

    thumb_filename = NEW(strlen(thumb_basename)+7, char);
    sprintf(thumb_filename, "%s.%ld",
	    thumb_basename, (docpages.num_pages % 99999) + 1L);
    thumb_ref = read_thumbnail(thumb_filename);
    RELEASE(thumb_filename);
    if (thumb_ref)
      pdf_add_dict(currentpage->page_obj, pdf_new_name("Thumb"), thumb_ref);
  }
#endif /* !NO_THUMBNAIL */

  docpages.num_pages++;
}


static void
fill_background (void)
{
  struct page_entry *currentpage;
  pdf_color  bgcolor;
  pdf_rect   bbox;
  int        colormode;
  pdf_obj   *saved_content;

  colormode = pdf_dev_get_param(PDF_DEV_PARAM_COLORMODE);
  if (!colormode) {
    return;
  }

  pdf_color_get_bgcolor(&bgcolor);
  if (pdf_color_is_white(&bgcolor)) {
    return;
  }

  pdf_doc_get_mediabox(pdf_doc_current_page_number(), &bbox);

  currentpage = LASTPAGE();
  ASSERT(currentpage);

  if (!currentpage->background)
    currentpage->background = pdf_new_stream(STREAM_COMPRESS);

  saved_content = currentpage->contents;
  currentpage->contents = currentpage->background;

  pdf_dev_gsave();

  pdf_color_push();
  pdf_dev_setcolor(&bgcolor, 1); /* is_fill */
  pdf_dev_moveto(bbox.llx, bbox.lly);
  pdf_dev_lineto(bbox.urx, bbox.lly);
  pdf_dev_lineto(bbox.urx, bbox.ury);
  pdf_dev_lineto(bbox.llx, bbox.ury);
  pdf_dev_closepath();
  pdf_dev_fill();
  pdf_color_pop();

  pdf_dev_grestore();

  currentpage->contents = saved_content;

  return;
}

void
pdf_doc_begin_page (double scale, double x_origin, double y_origin)
{
  pdf_tmatrix M;

  M.a = scale; M.b = 0.0;
  M.c = 0.0  ; M.d = scale;
  M.e = x_origin;
  M.f = y_origin;

  /* pdf_doc_new_page() allocates page content stream. */
  pdf_doc_new_page();

  pdf_dev_bop(&M);
}

/* ... */
#include "mpost.h"
void
pdf_doc_end_page (void)
{
  /* Finish any pending PS specials */
  mps_eop_cleanup();

  pdf_dev_eop();
  fill_background();

  pdf_doc_finish_page();
}

void
pdf_doc_add_page_content (const char *buffer, unsigned length)
{
  struct page_entry *currentpage;

  currentpage = LASTPAGE();

  ASSERT(currentpage && currentpage->contents);

  pdf_add_stream(currentpage->contents, buffer, length);
}

static char *doccreator = NULL; /* Ugh */

void
pdf_open_document (const char *filename,
		   double media_width, double media_height,
		   double annot_grow_amount, int bookmark_open_depth)
{
  pdf_out_init(filename);

  docroot.catalog = pdf_new_dict();
  pdf_set_root(docroot.catalog);

  annot_grow = annot_grow_amount;

  pdf_init_resources();
  pdf_init_colors();
  pdf_init_fonts();
  /* Thumbnail want this to be initialized... */
  pdf_init_images();

  docinfo = pdf_new_dict();
  if (doccreator) {
    pdf_add_dict(docinfo,
		 pdf_new_name("Creator"),
		 pdf_new_string(doccreator, strlen(doccreator)));
    RELEASE(doccreator);
    doccreator = NULL;
  }
  pdf_set_info(docinfo);

  doc_init_bookmarks(bookmark_open_depth);
  doc_init_articles ();
  doc_init_names    ();
  doc_init_page_tree(media_width, media_height);

  create_encrypt();

#ifndef NO_THUMBNAIL
  /* Create a default name for thumbnail image files */
  if (thumb_enabled) {
    if (strlen(filename) > 4 &&
	!strncmp(".pdf", filename + strlen(filename) - 4, 4)) {
      thumb_basename = NEW(strlen(filename)-4+1, char);
      strncpy(thumb_basename, filename, strlen(filename)-4);
      thumb_basename[strlen(filename)-4] = 0;
      } else {
	thumb_basename = NEW(strlen(filename)+1, char);
	strcpy(thumb_basename, filename);
      }
  }
#endif /* !NO_THUMBNAIL */
}

void
pdf_doc_set_creator (const char *creator)
{
  if (!creator ||
      creator[0] == '\0')
    return;

  doccreator = NEW(strlen(creator)+1, char);
  strcpy(doccreator, creator); /* Ugh */
}

static void
doc_close_catalog (void)
{
  pdf_obj *tmp;

  if (docroot.viewerpref) {
    tmp = pdf_lookup_dict(docroot.catalog, "ViewerPreferences");
    if (!tmp) {
      pdf_add_dict(docroot.catalog,
		   pdf_new_name("ViewerPreferences"), pdf_ref_obj(docroot.viewerpref));
    } else if (PDF_OBJ_DICTTYPE(tmp)) {
      pdf_merge_dict(docroot.viewerpref, tmp);
      pdf_add_dict(docroot.catalog,
		   pdf_new_name("ViewerPreferences"), pdf_ref_obj(docroot.viewerpref));
    } else { /* Maybe reference */
      /* What should I do? */
      WARN("Could not modify ViewerPreferences.");
    }
    pdf_release_obj(docroot.viewerpref);
    docroot.viewerpref = NULL;
  }

  if (docroot.pagelabels) {
    tmp = pdf_lookup_dict(docroot.catalog, "PageLabels");
    if (!tmp) {
      tmp = pdf_new_dict();
      pdf_add_dict(tmp, pdf_new_name("Nums"),  pdf_link_obj(docroot.pagelabels));
      pdf_add_dict(docroot.catalog,
		   pdf_new_name("PageLabels"), pdf_ref_obj(tmp));
      pdf_release_obj(tmp);
    } else { /* Maybe reference */
      /* What should I do? */
      WARN("Could not modify PageLabels.");
    }
    pdf_release_obj(docroot.pagelabels);
    docroot.pagelabels = NULL;
  }

  pdf_add_dict(docroot.catalog,
	       pdf_new_name("Type"), pdf_new_name("Catalog"));
  pdf_release_obj(docroot.catalog);
  docroot.catalog = NULL;
}

void
pdf_close_document (void)
{
  /*
   * Following things were kept around so user can add dictionary items.
   */
  doc_close_articles ();
  doc_close_names    ();
  doc_close_bookmarks();
  doc_close_page_tree();
  doc_close_docinfo  ();
  doc_close_catalog  ();

  pdf_close_images();
  pdf_close_fonts ();
  pdf_close_colors();

  pdf_close_resources(); /* Should be at last. */

  pdf_out_flush();

#ifndef NO_THUMBNAIL
  if (thumb_basename)
    RELEASE(thumb_basename);
#endif /* !NO_THUMBNAIL */
}

/*
 * All this routine does is give the form a name and add a unity scaling matrix.
 * It fills in required fields.  The caller must initialize the stream.
 */
void
pdf_doc_make_xform (pdf_obj     *xform,
		    pdf_rect    *bbox,
		    pdf_tmatrix *matrix,
		    pdf_obj     *resources)
{
  pdf_obj *xform_dict;
  pdf_obj *tmp;

  xform_dict = pdf_stream_dict(xform);
  pdf_add_dict(xform_dict,
	       pdf_new_name("Type"),     pdf_new_name("XObject"));
  pdf_add_dict(xform_dict,
	       pdf_new_name("Subtype"),  pdf_new_name("Form"));
  pdf_add_dict(xform_dict,
	       pdf_new_name("FormType"), pdf_new_number(1.0));

  if (!bbox)
    ERROR("No BoundingBox supplied.");

  tmp = pdf_new_array();
  pdf_add_array(tmp, pdf_new_number(ROUND(bbox->llx, .001)));
  pdf_add_array(tmp, pdf_new_number(ROUND(bbox->lly, .001)));
  pdf_add_array(tmp, pdf_new_number(ROUND(bbox->urx, .001)));
  pdf_add_array(tmp, pdf_new_number(ROUND(bbox->ury, .001)));
  pdf_add_dict(xform_dict, pdf_new_name("BBox"), tmp);

  if (matrix) {
    tmp = pdf_new_array();
    pdf_add_array(tmp, pdf_new_number(ROUND(matrix->a, .00001)));
    pdf_add_array(tmp, pdf_new_number(ROUND(matrix->b, .00001)));
    pdf_add_array(tmp, pdf_new_number(ROUND(matrix->c, .00001)));
    pdf_add_array(tmp, pdf_new_number(ROUND(matrix->d, .00001)));
    pdf_add_array(tmp, pdf_new_number(ROUND(matrix->e, .001  )));
    pdf_add_array(tmp, pdf_new_number(ROUND(matrix->f, .001  )));
    pdf_add_dict(xform_dict, pdf_new_name("Matrix"), tmp);
  }

  pdf_add_dict(xform_dict, pdf_new_name("Resources"), resources);
}

struct resource_stack 
{
  int      xform_depth;

  double   ref_x;
  double   ref_y;
  pdf_rect cropbox;
  pdf_obj *contents;
  pdf_obj *resources;
} res_stack[PDF_DOC_GRABBING_NEST_MAX];

/*
 * begin_form_xobj creates an xobject with its "origin" at
 * xpos and ypos that is clipped to the specified bbox. Note
 * that the origin is not the lower left corner of the bbox.
 */
int
pdf_doc_begin_grabbing (const char *ident,
			double ref_x, double ref_y, const pdf_rect *cropbox)
{
  int                xobj_id = -1;
  struct page_entry *currentpage;
  xform_info         info;

  if (xobjects_pending >= PDF_DOC_GRABBING_NEST_MAX) {
    WARN("Form XObjects nested too deeply. (limit: %d)", PDF_DOC_GRABBING_NEST_MAX);
    return -1;
  }
  /*
   * This is a real hack.  We basically treat each xobj as a separate mini
   * page unto itself.  Save all the page structures and reinitialize
   * them when we finish this xobject.
   */
  currentpage = LASTPAGE();
  res_stack[xobjects_pending].xform_depth = pdf_dev_current_depth();
  res_stack[xobjects_pending].ref_x = currentpage->ref_x;
  res_stack[xobjects_pending].ref_y = currentpage->ref_y;
  res_stack[xobjects_pending].cropbox.llx = currentpage->cropbox.llx;
  res_stack[xobjects_pending].cropbox.lly = currentpage->cropbox.lly;
  res_stack[xobjects_pending].cropbox.urx = currentpage->cropbox.urx;
  res_stack[xobjects_pending].cropbox.ury = currentpage->cropbox.ury;
  res_stack[xobjects_pending].contents  = currentpage->contents;
  res_stack[xobjects_pending].resources = currentpage->resources;

  currentpage->ref_x = ref_x;
  currentpage->ref_y = ref_y;
  currentpage->cropbox.llx = cropbox->llx;
  currentpage->cropbox.lly = cropbox->lly;
  currentpage->cropbox.urx = cropbox->urx;
  currentpage->cropbox.ury = cropbox->ury;

  currentpage->contents  = pdf_new_stream(STREAM_COMPRESS);
  currentpage->resources = pdf_new_dict();

  pdf_ximage_init_form_info(&info);

  info.matrix.a = 1.0; info.matrix.b = 0.0;
  info.matrix.c = 0.0; info.matrix.d = 1.0;
  info.matrix.e = -ref_x;
  info.matrix.f = -ref_y;

  info.bbox.llx = cropbox->llx;
  info.bbox.lly = cropbox->lly;
  info.bbox.urx = cropbox->urx;
  info.bbox.ury = cropbox->ury;

  /* Use reference since content itself isn't available yet. */
  xobj_id = pdf_ximage_defineresource(ident,
				      PDF_XOBJECT_TYPE_FORM,
				      &info, pdf_ref_obj(currentpage->contents));

  /*
   * Make sure the object is self-contained by adding the
   * current font and color to the object stream.
   */
  pdf_dev_reset();

  xobjects_pending++;

  return xobj_id;
}

void
pdf_doc_end_grabbing (void)
{
  pdf_obj    *xform;
  pdf_obj    *tmp;
  pdf_tmatrix matrix;
  struct page_entry *currentpage;

  if (xobjects_pending <= 0) {
    WARN("[doc] end_form_xobj: Tried to close a nonexistent xobject.");
    return;
  }

  currentpage = LASTPAGE();

  xform = currentpage->contents;
  /*
   * ProcSet is obsolete in PDF-1.4 but recommended for compatibility.
   */
  tmp = pdf_new_array();
  pdf_add_array(tmp, pdf_new_name("PDF"));
  pdf_add_array(tmp, pdf_new_name("Text"));
  pdf_add_array(tmp, pdf_new_name("ImageC"));
  pdf_add_array(tmp, pdf_new_name("ImageB"));
  pdf_add_array(tmp, pdf_new_name("ImageI"));
  pdf_add_dict(currentpage->resources, pdf_new_name("ProcSet"), tmp);

  /*
   * The reference point of an Xobject is at the lower left corner
   * of the bounding box.  Since we would like to have an arbitrary
   * reference point, we use a transformation matrix, translating
   * the reference point to (0,0).
   */

  matrix.a = 1.0; matrix.b = 0.0;
  matrix.c = 0.0; matrix.d = 1.0;
  matrix.e = -(currentpage->ref_x);
  matrix.f = -(currentpage->ref_y);

  pdf_doc_make_xform(xform,
		     &(currentpage->cropbox), &matrix,
		     pdf_ref_obj(currentpage->resources));
  pdf_release_obj(currentpage->resources);
  pdf_release_obj(xform);

  currentpage = LASTPAGE();

  xobjects_pending--;

  pdf_dev_grestore_to(res_stack[xobjects_pending].xform_depth);

  currentpage->ref_x       = res_stack[xobjects_pending].ref_x;
  currentpage->ref_y       = res_stack[xobjects_pending].ref_y;
  currentpage->cropbox.llx = res_stack[xobjects_pending].cropbox.llx;
  currentpage->cropbox.lly = res_stack[xobjects_pending].cropbox.lly;
  currentpage->cropbox.urx = res_stack[xobjects_pending].cropbox.urx;
  currentpage->cropbox.ury = res_stack[xobjects_pending].cropbox.ury;
  currentpage->contents    = res_stack[xobjects_pending].contents;
  currentpage->resources   = res_stack[xobjects_pending].resources;

  pdf_dev_reset();

  return;
}

static struct
{
  int      dirty;
  pdf_obj *annot_dict;
  pdf_rect rect;
} breaking_state = {0, NULL, {0.0, 0.0, 0.0, 0.0}};

static void
reset_box (void)
{
  breaking_state.rect.llx = 10000.0; /* large value */
  breaking_state.rect.lly = 10000.0; /* large value */
  breaking_state.rect.urx = 0.0;     /* small value */
  breaking_state.rect.ury = 0.0;     /* small value */
  breaking_state.dirty    = 0;
}

void
pdf_doc_begin_annot (pdf_obj *dict)
{
  breaking_state.annot_dict = dict;
  reset_box();
}

void
pdf_doc_end_annot (void)
{
  pdf_doc_break_annot();
  breaking_state.annot_dict = NULL;
}

void
pdf_doc_break_annot (void)
{
  if (breaking_state.dirty) {
    pdf_obj  *annot_dict;

    /* Copy dict */
    annot_dict = pdf_new_dict();
    pdf_merge_dict(annot_dict, breaking_state.annot_dict);
    pdf_doc_add_annot(pdf_doc_current_page_number(),
		      &(breaking_state.rect), annot_dict);
    pdf_release_obj(annot_dict);
  }
  reset_box();
}

void
pdf_doc_expand_box (const pdf_rect *rect)
{
  breaking_state.rect.llx = MIN(breaking_state.rect.llx, rect->llx);
  breaking_state.rect.lly = MIN(breaking_state.rect.lly, rect->lly);
  breaking_state.rect.urx = MAX(breaking_state.rect.urx, rect->urx);
  breaking_state.rect.ury = MAX(breaking_state.rect.ury, rect->ury);
  breaking_state.dirty    = 1;
}

/* This should be number tree */
void
pdf_doc_set_pagelabel (long pg_start,
		       const char *type,
		       const void *prefix, int prfx_len, long start)
{
  pdf_obj *label_dict;

  if (!docroot.pagelabels)
    docroot.pagelabels = pdf_new_array();

  label_dict = pdf_new_dict();
  if (!type || type[0] == '\0') /* Set back to default. */
    pdf_add_dict(label_dict, pdf_new_name("S"),  pdf_new_name("D"));
  else {
    if (type)
      pdf_add_dict(label_dict, pdf_new_name("S"), pdf_new_name(type));
    if (prefix && prfx_len > 0)
      pdf_add_dict(label_dict,
		   pdf_new_name("P"),
		   pdf_new_string(prefix, prfx_len));
    if (start != 1)
      pdf_add_dict(label_dict,
		   pdf_new_name("St"), pdf_new_number(start));
  }

  pdf_add_array(docroot.pagelabels, pdf_new_number(pg_start));
  pdf_add_array(docroot.pagelabels, label_dict);
}
